Package: profmem
Version: 0.7.0
Title: Simple Memory Profiling for R
Imports: utils
Suggests: R.rsp, markdown, microbenchmark
VignetteBuilder: R.rsp
Authors@R: c(person("Henrik", "Bengtsson", role=c("aut", "cre", "cph"),
                                          email = "henrikb@braju.com"))
Description: A simple and light-weight API for memory profiling of R expressions.  The profiling is built on top of R's built-in memory profiler ('utils::Rprofmem()'), which records every memory allocation done by R (also native code).
License: LGPL (>= 2.1)
LazyLoad: TRUE
URL: https://henrikbengtsson.github.io/profmem/,
        https://github.com/HenrikBengtsson/profmem
BugReports: https://github.com/HenrikBengtsson/profmem/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2025-05-02 20:14:48 UTC; henrik
Author: Henrik Bengtsson [aut, cre, cph]
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2025-05-02 21:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,Rprofmem)
S3method(c,Rprofmem)
S3method(print,Rprofmem)
S3method(subset,Rprofmem)
S3method(total,Rprofmem)
export(profmem)
export(profmem_begin)
export(profmem_depth)
export(profmem_end)
export(profmem_resume)
export(profmem_status)
export(profmem_suspend)
export(readRprofmem)
export(total)
importFrom(utils,Rprofmem)
importFrom(utils,file_test)

#' Total number of bytes allocated
#'
#' @param x An `Rprofmem` object.
#' @param ... Not used.
#'
#' @return A non-negative numeric.
#'
#' @aliases total.Rprofmem subset.Rprofmem
#' @export
#' @keywords internal
total <- function(x, ...) UseMethod("total")

#' @export
total.Rprofmem <- function(x, ...) {
  sum(x$bytes, na.rm=TRUE)
}

#' @export
c.Rprofmem <- function(...) {
  args <- list(...)
  
  what <- NULL
  bytes <- NULL
  trace <- NULL
  threshold <- NULL
  
  for (arg in args) {
    stop_if_not(inherits(arg, "Rprofmem"))
    what <- c(what, arg$what)
    bytes <- c(bytes, arg$bytes)
    trace <- c(trace, arg$trace)
    threshold <- c(threshold, attr(arg, "threshold"))
  }
  threshold <- max(threshold)
  stop_if_not(length(threshold) == 1, is.finite(threshold),
            is.integer(threshold), threshold >= 0L)
  
  res <- data.frame(what = what, bytes = bytes, stringsAsFactors = FALSE)
  res$trace <- trace
  if (threshold > 0L) {
    keep <- is.na(bytes) | (bytes >= threshold)
    stop_if_not(all(is.finite(keep)))
    res <- res[keep, ]
  }
  attr(res, "threshold") <- threshold
  class(res) <- c("Rprofmem", class(res))
  ## Sanity check
  stop_if_not(all(c("what", "bytes", "trace") %in% names(res)))
  
  res
}

#' @export
subset.Rprofmem <- function(x, ...) {
  res <- NextMethod()
  attr(res, "expression") <- attr(x, "expression")
  attr(res, "threshold") <- attr(x, "threshold")
  res
}

#' @export
as.data.frame.Rprofmem <- function(x, ...) {
  what <- x$what
  bytes <- x$bytes
  traces <- unlist(lapply(x$trace, FUN=function(x) {
    trace <- rev(x)
    hasName <- !grepl("^<[^>]*>$", trace)
    trace[hasName] <- sprintf("%s()", trace[hasName])
    paste(trace, collapse=" -> ")
  }))
  
  res <- data.frame(what = what, bytes = bytes, calls = traces,
                    stringsAsFactors = FALSE)

  ## Preserve row names
  rownames(res) <- rownames(x)
  
  res
} ## as.data.frame()


#' @export
print.Rprofmem <- function(x, expr = getOption("profmem.print.expr", TRUE), newpage = getOption("profmem.print.newpage", FALSE), calls = getOption("profmem.print.calls", TRUE), ...) {
  if (expr && "expression" %in% names(attributes(x))) {
    cat("Rprofmem memory profiling of:\n")
    print(attr(x, "expression"))
    cat("\n")
  }

  threshold <- attr(x, "threshold")
  if (is.null(threshold) || threshold == 0L) {
    cat("Memory allocations:\n")
  } else {
    cat(sprintf("Memory allocations (>= %g bytes):\n", threshold))
  }

  data <- as.data.frame(x, ..., stringsAsFactors = FALSE)
  
  if (!newpage) {
    drop <- which(x$what == "new page")
    if (length(drop) > 0) {
      cat(sprintf("Number of 'new page' entries not displayed: %d\n",
                  length(drop)))
      data <- data[-drop, ]
    }
  }
  
  n <- nrow(data)
  total <- sum(data$bytes, na.rm=TRUE)

  ## Number of digits for indices
  widx <- floor(log10(n)+1)

  ## Number of digits for bytes
  wbytes <- floor(log10(total)+1)

  ## Report empty call stack as "internal"
  data$calls[!nzchar(data$calls)] <- "<internal>"

  data <- rbind(data, list(what = "", bytes = total, calls = ""))
  rownames(data)[n+1] <- "total"

  if (!calls) {
    data$calls <- NULL
  }

  print(data, ...)

  ## Any errors to report on?
  error <- attr(x, "error")
  if (!is.null(error)) {
    cat(sprintf("\nNote, an error occurred while evaluating the expression: %s\n", conditionMessage(error)))
  }

  invisible(x)
} ## print()

#' Read an Rprofmem log file
#'
#' Reads and parses an Rprofmem log file that was created by
#' [utils::Rprofmem()].
#'
#' @param pathname The Rprofmem log file to be read.
#' 
#' @param as Specifies in what format data should be returned.
#' If `"raw"`, the line content of the file is returned as is
#' (as a character vector).
#' If `"fixed"`, as `"raw"` but with missing newlines added to lines with empty
#' stack calls that may be introduced in R (< 3.5.0) (see Ref. 1).
#' If `"Rprofmem"`, the collected Rprofmem data is fully
#' parsed into bytes and call stack information.
#' 
#' @param drop Number of levels to drop from the top of the call stack.
#' 
#' @param ... Not used
#'
#' @return An `Rprofmem` data.frame or a character vector (if `as` is `"raw"`
#' or `"fixed"`).
#' An `Rprofmem` data.frame has columns `what`, `bytes`, and `trace`, with:
#'
#'  * `what`:  (character) type of memory event;
#'             either `"alloc"` or `"new page"`
#'  * `bytes`: (numeric) number of bytes allocated or `NA_real_`
#'             (when `what` is `"new page"`)
#'  * `trace`: (list of character vectors) zero or more function names
#'
#' @example incl/readRprofmem.R
#'
#' @references
#' Ref. 1: \url{https://github.com/HenrikBengtsson/Wishlist-for-R/issues/25}
#'
#' @export
#' @importFrom utils file_test
readRprofmem <- function(pathname, as = c("Rprofmem", "fixed", "raw"), drop = 0L, ...) {
  stop_if_not(file_test("-f", pathname))
  as <- match.arg(as)
  drop <- as.integer(drop)
  stop_if_not(length(drop) == 1, drop >= 0)

  ## Read raw
  bfr <- readLines(pathname, warn=FALSE)
  if (as == "raw") return(bfr)


  ## WORKAROUND: Add newlines for entries with empty call stacks
  ## https://github.com/HenrikBengtsson/Wishlist-for-R/issues/25
  pattern <- "^(new page|[0-9]+)[ ]?:(new page|[0-9]+)[ ]?:"
  while(any(grepl(pattern, bfr))) {
    bfr <- gsub(pattern, "\\1 :\n\\2 :", bfr)
    bfr <- unlist(strsplit(bfr, split="\n", fixed=TRUE))
  }

  if (as == "fixed") return(bfr)

  if (getOption("profmem.debug", FALSE)) print(bfr)
  
  ## Parse Rprofmem results
  pattern <- "^([0-9]+|new page)[ ]?:(.*)"
  res <- lapply(bfr, FUN=function(x) {
    bytes <- gsub(pattern, "\\1", x)
    what <- rep("alloc", times = length(x))
    idxs <- which(bytes == "new page")
    if (length(idxs) > 0) {
      what[idxs] <- "new page"
      bytes[idxs] <- ""  # Will become NA below w/out warning
    }
    bytes <- as.numeric(bytes)

    trace <- gsub(pattern, "\\2", x)
    trace <- gsub('" "', '", "', trace, fixed=TRUE)
    trace <- sprintf("c(%s)", trace)
  
    trace <- eval(parse(text=trace), enclos = baseenv())
    trace <- trace[seq_len(max(0L, length(trace)-drop))]

    list(what = what, bytes = bytes, trace = trace)
  })

  if (length(res) == 0) {
    what <- character(0L)
    bytes <- numeric(0L)
    traces <- list()
  } else {
    what <- unlist(lapply(res, FUN=function(x) x$what), use.names=FALSE)
    bytes <- unlist(lapply(res, FUN=function(x) x$bytes), use.names=FALSE)
    traces <- lapply(res, FUN=function(x) x$trace)
  }
  res <- data.frame(what = what, bytes = bytes, stringsAsFactors = FALSE)
  res$trace <- traces
  bfr <- bytes <- traces <- NULL
  
  class(res) <- c("Rprofmem", class(res))

  ## Sanity check
  stop_if_not(all(c("what", "bytes", "trace") %in% names(res)))
  
  res
} ## readRprofmem()

library("profmem")

profmem_stack <- profmem:::profmem_stack

## In case there are active profmem session in the current R session
## (shouldn't happen during 'R CMD check' though)
while (profmem_stack("depth") > 0) profmem_stack("pop")

## Trying to pop from empty stack
res <- tryCatch({
  p <- profmem_stack("pop")
}, error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Trying non-existing stack action
res <- tryCatch({
  p <- profmem_stack("non-existing")
}, error = identity)
print(res)
stopifnot(inherits(res, "error"))


if (capabilities("profmem")) {

## Trying to end existing profmem session
res <- tryCatch({
  p <- profmem_end()
}, error = identity)
print(res)
stopifnot(inherits(res, "error"))

} ## if (capabilities("profmem"))

# A lookup-only-once version of capabilities("profmem")
# Benchmarking shows it's 8-10 times faster this way.
capableOfProfmem <- local({
  res <- NA
  function() {
    if (is.na(res)) {
      res <<- capabilities("profmem")
    }
    res
  }
})

library("profmem")

message("readRprofmem() ...")

message(" - corrupt file")

broken <- system.file("extdata", "broken.Rprofmem.out", package = "profmem")

bfr <- readLines(broken)
cat("readLines(broken):\n")
print(bfr)

raw <- readRprofmem(broken, as = "raw")
cat("readRprofmem(broken, as = 'raw'):\n")
print(raw)
stopifnot(
  length(raw) == length(bfr),
  all(raw == bfr)
)

message("readRprofmem(broken, as = 'fixed'):\n")
fixed <- readRprofmem(broken, as = "fixed")
print(fixed)
stopifnot(length(fixed) >= length(bfr))

p <- readRprofmem(broken, as = "Rprofmem")
cat("readRprofmem(broken, as = 'Rprofmem'):\n")
print(p)
str(p)
stopifnot(nrow(p) == length(fixed))


message(" - empty file")

options(profmem.debug = TRUE)

empty <- tempfile()
writeLines(character(0L), con = empty)

bfr <- readLines(empty)
stopifnot(length(bfr) == 0)

raw <- readRprofmem(empty, as = "raw")
stopifnot(is.character(raw), length(raw) == 0)

fixed <- readRprofmem(empty, as = "fixed")
stopifnot(is.character(raw), length(raw) == 0)

p <- readRprofmem(empty, as = "Rprofmem")
stopifnot(nrow(p) == 0L)

options(profmem.debug = FALSE)


if (capabilities("profmem")) {
  
  live <- tempfile()
  Rprofmem(live)
  x <- raw(1000)
  A <- matrix(rnorm(100), ncol=10)
  Rprofmem()

  bfr <- readLines(live)
  cat("readLines(live):\n")
  print(bfr)
  
  raw <- readRprofmem(live, as = "raw")
  cat("readRprofmem(live, as = 'raw'):\n")
  print(raw)
  stopifnot(
    length(raw) == length(bfr),
    all(raw == bfr)
  )

  fixed <- readRprofmem(live, as = "fixed")
  cat("readRprofmem(live, as = 'fixed'):\n")
  print(fixed)
  stopifnot(length(fixed) >= length(bfr))

  p <- readRprofmem(live, as = "Rprofmem")
  cat("readRprofmem(live, as = 'Rprofmem'):\n")
  print(p)
  str(p)
  stopifnot(nrow(p) == length(fixed))
  
} ## if (capabilities("profmem"))

message("readRprofmem() ... DONE")

truth <- capabilities("profmem")
print(truth)

cache <- profmem:::capableOfProfmem()
print(cache)
stopifnot(identical(cache, truth))

cache <- profmem:::capableOfProfmem()
print(cache)
stopifnot(identical(cache, truth))

## Fake calling profmem() when memory profiling is disabled
## and assert that profmem() throws an error
f <- profmem:::capableOfProfmem
environment(f)$res <- FALSE
t <- profmem:::capableOfProfmem()
print(t)
stopifnot(identical(t, FALSE))
res <- tryCatch({
  p <- profmem::profmem(x <- 1:1000)
}, error = identity)
stopifnot(inherits(res, "simpleError"))
environment(f)$res <- cache  ## Undo

stop_if_not <- function(...) {
  res <- list(...)
  n <- length(res)
  if (n == 0L) return()

  for (ii in 1L:n) {
    res_ii <- .subset2(res, ii)
    if (length(res_ii) != 1L || is.na(res_ii) || !res_ii) {
        mc <- match.call()
        call <- deparse(mc[[ii + 1]], width.cutoff = 60L)
        if (length(call) > 1L) call <- paste(call[1L], "...")
        stop(sQuote(call), " is not TRUE", call. = FALSE, domain = NA)
    }
  }
}

#' Memory profiling R
#'
#' `profmem()` evaluates and memory profiles an \R expression.
#'
#' @param expr An \R expression to be evaluated and profiled.
#' 
#' @param envir The environment in which the expression should be evaluated.
#' 
#' @param substitute Should `expr` be [base::substitute()]:d or not.
#' 
#' @param threshold The smallest memory allocation (in bytes) to log.
#'
#' @param on_error (character) Controls whether evaluation errors should
#' signal an error (`"error"`), a warning (`"warning"), or be ignored
#' (`"ignore"`; default).
#'
#' @return `profmem()` and `profmem_end()` returns the collected allocation
#' data as an `Rprofmem` data.frame with additional attributes set.
#' An `Rprofmem` data.frame has columns `what`, `bytes`, and `trace`, with:
#'
#'  * `what`:  (character) type of memory event;
#'             either `"alloc"` or `"new page"`
#'  * `bytes`: (numeric) number of bytes allocated or `NA_real_`
#'             (when `what` is `"new page"`)
#'  * `trace`: (list of character vectors) zero or more function names
#'
#' The attributes set are:
#'
#'  * `threshold` : The threshold used (= argument `threshold`)
#'  * `expression`: The expression profiled (= argument `expr`)
#'  * `value`     : The value of the evaluated expression
#'                  (only set if there was no error)
#'  * `error`     : The error object in case the evaluation failed
#'                  (only set if there was an error)
#'
#' @details
#' In order for memory profiling to work, \R must have been _built_ with memory
#' profiling enabled.  Function
#' \code{\link[base:capabilities]{base::capabilities("profmem")}} will
#' return `TRUE` of it is enabled, otherwise `FALSE`.
#' If memory profiling is _not_ supported, `profmem()` and `profmem_begin()`
#' will produce an informative error.  The pre-built \R binaries on
#' [CRAN](https://cran.r-project.org/) support memory profiling.
#'
#' What is logged?  The `profmem()` function uses [utils::Rprofmem()] for
#' logging memory, which logs all memory _allocations_ that are done via the
#' \R framework.  Specifically, the logger is tied to `allocVector3()` part
#' of \R's native API.  This means that nearly all memory allocations done
#' in \R are logged. _Neither_ memory deallocations _nor_ garbage collection
#' events are logged.  Furthermore, allocations done by non-\R native libraries
#' or \R packages that use native code `Calloc() / Free()` for internal objects
#' are also _not_ logged.
#'
#' Any memory events that would occur due to calling any of the \pkg{profmem}
#' functions themselves will _not_ be logged and _not_ be part of the returned
#' profile data (regardless whether memory profiling is active or not).
#' This is intentional.
#'
#' If a profmem profiling is already active, `profmem()` and `profmem_begin()`
#' performs an _independent_, _nested_ profiling, which does not affect the
#' already active one.  When the active one completes, it will contain all
#' memory events also collected by the nested profiling as if the nested one
#' never occurred.
#'
#' Profiling gathered by \pkg{profmem} _will_ be corrupted if the code profiled
#' calls [utils::Rprofmem()], with the exception of such calls done via the
#' \pkg{profmem} package itself.
#'
#'
#' @example incl/profmem.R
#'
#' @export
#' @importFrom utils Rprofmem
profmem <- function(expr, envir = parent.frame(), substitute = TRUE, threshold = getOption("profmem.threshold", 0L), on_error = c("ignore", "warning", "error")) {
  if (substitute) expr <- substitute(expr)
  on_error <- match.arg(on_error, choices = c("ignore", "warning", "error"))

  profmem_begin(threshold = threshold)

  ## Profile memory
  error <- NULL
  value <- tryCatch({
    eval(expr, envir=envir, enclos = baseenv())
  }, error = function(ex) {
    error <<- ex
    NULL
  })

  pm <-  profmem_end()

  if (on_error != "ignore") {
    action <- switch(on_error, error = stop, warning = warning)
    action(paste("profmem::profmem() detected a run-time error:",
           conditionMessage(error)))
  }

  ## Annotate
  attr(pm, "expression") <- expr
  attr(pm, "value") <- value
  attr(pm, "error") <- error

  pm
} ## profmem()


#' Memory profiling R
#'
#' `profmem_begin()` starts the memory profiling of all the following \R
#' evaluations until `profmem_end()` is called.
#'
#' @return `profmem_begin()` returns (invisibly) the number of nested profmem
#' session currently active.
#' 
#' @rdname profmem
#' @export
profmem_begin <- function(threshold = getOption("profmem.threshold", 0L)) {
  ## Is memory profiling supported?
  if (!capableOfProfmem()) {
    msg <- "Profiling of memory allocations is not supported on this R system (capabilities('profmem') reports FALSE). See help('tracemem')."
    if (.Platform$OS.type == "unix") {
      msg <- paste(msg, "To enable memory profiling for R on Linux, R needs to be configured and built using './configure --enable-memory-profiling'.")
    }
    stop(msg)
  }

  threshold <- as.integer(threshold)
  stop_if_not(length(threshold) == 1, is.finite(threshold), threshold >= 0L)

  depth <- profmem_stack("depth")
  if (depth > 0) {
    threshold_parent <- profmem_stack("threshold")
    if (threshold > threshold_parent) {
      warning(sprintf("Nested profmem threshold (%d bytes) cannot be greater than the threshold (%d bytes) of active profmem sessions (n = %d). Will use the active threshold instead.", threshold, threshold_parent, depth))
      threshold <- threshold_parent
    }
  }
  
  profmem_suspend()
   
  ## Push new level
  depth <- profmem_stack("push", threshold = threshold)
 
  profmem_resume()
 
  invisible(depth)
}

#' @rdname profmem
#' @export
profmem_end <- function() {
  profmem_suspend()

  depth <- profmem_stack("depth")
  if (depth == 0) {
    stop("Did you forget to call profmem_begin()?")
  }

  data <- profmem_stack("pop")

  profmem_resume()
  
  data
}

#' `profmem_suspend()` suspends an active profiling until resumed by
#' `profmem_resume()` or ended by `profmem_end()`.
#' Calling `profmem_begin()` or `profmem()` will resume any suspended
#' profiling. _Nested_ resuming and suspending is _not_ supported;
#' it is a global state.
#' 
#' @return `profmem_suspend()` and `profmem_resume()` returns nothing.
#'
#' @rdname profmem
#' @importFrom utils Rprofmem
#' @export
profmem_suspend <- function() {
  ## Works regardless of active Rprofmem exists or not
  Rprofmem("")

  profmem_stack("suspend")
  
  ## Nothing more to do?
  if (profmem_stack("depth") == 0) return()
  
  ## Import current log
  drop <- length(sys.calls()) + 4L
  pathname <- profmem_pathname()
  data <- readRprofmem(pathname, drop = drop)
  attr(data, "threshold") <- profmem_stack("threshold")
  profmem_stack("append", data)

  invisible()
}

#' @rdname profmem
#' @importFrom utils Rprofmem
#' @export
profmem_resume <- function() {
  threshold <- profmem_stack("threshold")
  pathname <- profmem_pathname()
  Rprofmem(filename = pathname, threshold = threshold)
  profmem_stack("resume")
  invisible()
}


#' `profmem_status()` checks whether there is an active profmem session
#' or not, and whether it is suspended or not.
#' 
#' @return `profmem_status()` returns `"inactive"`, `"active"`,
#' or `"suspended"`.
#'
#' @rdname profmem
#' @export
profmem_status <- function() {
  profmem_stack("status")
}

#' `profmem_depth()` gets the number of nested / stacked profmem sessions.
#'
#' @return `promem_depth()` returns a non-negative integer.
#'
#' @rdname profmem
#' @export
profmem_depth <- function() {
  profmem_stack("depth")
}



profmem_pathname <- local({
  pathname <- NULL
  function() {
    if (!is.null(pathname)) return(pathname)
    pathname <<- tempfile(pattern = "profmem.", fileext = ".Rprofmem.out")
    pathname
  }
})

profmem_stack <- local({
  empty <- data.frame(what = NULL, bytes = NULL, trace = NULL, stringsAsFactors = FALSE)
  empty <- structure(empty, class = c("Rprofmem", "data.frame"), threshold = 0L)
  
  stack <- list()
  suspended <- FALSE
  
  function(action = c("status", "depth", "threshold", "push", "pop", "append", "suspend", "resume"),
           data = empty, threshold = 0L) {
    action <- match.arg(action)

    ## Status queries
    if (action == "status") {
      if (length(stack) == 0) return("inactive")
      if (suspended) return("suspended")
      return("active")
    } else if (action == "depth") {
      return(length(stack))
    } else if (action == "threshold") {
      if (length(stack) == 0) return(NA_integer_)
      threshold <- attr(stack[[1]], "threshold")
      return(threshold)
    }
    
    ## State changing
    if (action == "suspend") {
      suspended <<- TRUE
      ## WORKAROUND: If not, above won't happen /2018-01-29
      dummy <- suspended
      return(invisible(suspended))
    } else if (action == "resume") {
      suspended <<- FALSE
      ## WORKAROUND: If not, above won't happen /2018-01-29
      dummy <- suspended
      return(invisible(suspended))
    }

    
    ## Stack changing
    if (action == "push") {
      stop_if_not(inherits(data, "Rprofmem"),
                length(threshold) == 1, is.finite(threshold),
                is.integer(threshold), threshold >= 0L)
      attr(data, "threshold") <- threshold
      stack <<- c(stack, list(data))
#      message("PUSH: stack depth: ", length(stack))
      return(length(stack))
    } else if (action == "pop") {
      depth <- length(stack)
      if (depth == 0) stop("Cannot 'pop' - profmem stack is empty")
      value <- stack[[depth]]
      stack <<- stack[-depth]
      depth <- length(stack)
      if (depth >= 1) {
        tmp <- stack[[depth]]
        tmp <- if (is.null(tmp)) value else c(tmp, value)
        stack[[depth]] <<- tmp
      }
#      message("POP: stack depth: ", length(stack))
      return(value)
    } else if (action == "append") {
      depth <- length(stack)
      if (depth == 0) stop("Cannot 'append' - profmem stack is empty")
      stop_if_not(inherits(data, "Rprofmem"),
                length(threshold) == 1, is.finite(threshold),
                is.integer(threshold), threshold >= 0L)
      attr(data, "threshold") <- threshold
      value <- stack[[depth]]
      value <- if (is.null(value)) data else c(value, data)
      stack[[depth]] <<- value
#      message("APPEND: stack depth: ", length(stack))
      return(invisible(value))
    }
  }
})



### TODO:
###
### profmem_add_string <- function(msg, ...) {
###   pathname <- profmem_pathname()
###   cat(msg, file = pathname, append = file_test("-f", pathname))
### }
### 
### ## FIXME: This produces lots of extra memory allocations, which
### ## we don't want to inject.
### profmem_add_note <- function(..., timestamp = TRUE) {
###   msg <- sprintf(...)
###   if (timestamp) {
###     msg <- sprintf("[%s] %s", format(Sys.time(), "%Y%m%d-%H%M%S"), msg)
###   }
###   msg <- sprintf("# %s\n", msg)
###   profmem_add_string(msg)
### }

library("profmem")

if (capabilities("profmem")) {

## Defaults to on_error = "ignore")
p <- profmem({
  x <- raw(1000)
  stop("Woops!")
})
print(p)


for (on_error in c("ignore", "warning", "error")) {
  message(sprintf("- profmem(..., on_error = \"%s\")", on_error))
  res <- tryCatch({
    profmem({
      x <- raw(1000)
      stop("Woops!")
    }, on_error = on_error)
  }, error = identity, warning = identity)
  if (on_error == "ignore") {
    stopifnot(inherits(res, "Rprofmem"))
  } else {
    stopifnot(inherits(res,  on_error))
  }
}

} ## if (capabilities("profmem"))

library("profmem")

message("profmem() - nested ...")

if (capabilities("profmem")) {

p1 <- profmem({
  x <- integer(1000)
  p2 <- profmem({
    Y <- matrix(x, nrow = 100, ncol = 10)
  })
  z <- Y + (x + 1)
})
print(p1)
print(p2)

p1 <- profmem({
  x <- integer(1000)
  p2 <- profmem({
    Y <- matrix(x, nrow = 100, ncol = 10)
  })
  z <- Y + (x + 1)
}, threshold = 500L)
print(p1)
print(p2)
  
## Cannot set a higher threshold than already active
p1 <- profmem({ p2 <- profmem({ }, threshold = 1000L) })

p1 <- profmem({
  ## Cannot set a higher threshold than already active
  res <- tryCatch({
    p2 <- profmem({ }, threshold = 1000L)
  }, warning = identify)
  stopifnot(inherits(res, "warning"))
})
print(p1)
print(p2)
 
}

message("profmem() - nested ... DONE")

library("profmem")

status <- profmem_status()
message("status: ", status)

depth <- profmem_depth()
message("depth: ", depth)

if (capabilities("profmem")) {

  p <- profmem({
    x <- raw(1000)
    A <- matrix(rnorm(100), ncol=10)
    status_2 <- profmem_status()
    depth_2 <- profmem_depth()
    message("depth_2: ", depth_2)
    stopifnot(depth2 == depth_+ 1)
  })
  print(p)

  data <- as.data.frame(p, stringsAsFactors = FALSE)
  print(data)

  t <- total(p)
  print(t)
  stopifnot(t == sum(data$bytes, na.rm=TRUE))


  foo <- function(n) {
    matrix(rnorm(n^2), ncol=n)
  }

  p <- profmem({
    A <- matrix(rnorm(100^2), ncol=100)
    B <- foo(100)
  })
  print(p)

  p2 <- subset(p, bytes >= 40000)
  print(p2)

  p <- profmem({
    A <- matrix(rnorm(100^2), ncol=100)
    B <- foo(100)
  }, threshold = 40000)
  d <- as.data.frame(p, stringsAsFactors = FALSE)
  print(d)

  p1 <- subset(p, !is.na(bytes))
  d1 <- as.data.frame(p1, stringsAsFactors = FALSE)
  
  p2 <- subset(p, bytes >= 40000)
  d2 <- as.data.frame(p2, stringsAsFactors = FALSE)
  print(d2)
  stopifnot(identical(d2, d1))

  print(dim(p))
  stopifnot(identical(dim(p), dim(d)))
  
} ## if (capabilities("profmem"))

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Rprofmem-class.R
\name{total}
\alias{total}
\alias{total.Rprofmem}
\alias{subset.Rprofmem}
\title{Total number of bytes allocated}
\usage{
total(x, ...)
}
\arguments{
\item{x}{An \code{Rprofmem} object.}

\item{...}{Not used.}
}
\value{
A non-negative numeric.
}
\description{
Total number of bytes allocated
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profmem.R
\name{profmem}
\alias{profmem}
\alias{profmem_begin}
\alias{profmem_end}
\alias{profmem_suspend}
\alias{profmem_resume}
\alias{profmem_status}
\alias{profmem_depth}
\title{Memory profiling R}
\usage{
profmem(
  expr,
  envir = parent.frame(),
  substitute = TRUE,
  threshold = getOption("profmem.threshold", 0L),
  on_error = c("ignore", "warning", "error")
)

profmem_begin(threshold = getOption("profmem.threshold", 0L))

profmem_end()

profmem_suspend()

profmem_resume()

profmem_status()

profmem_depth()
}
\arguments{
\item{expr}{An \R expression to be evaluated and profiled.}

\item{envir}{The environment in which the expression should be evaluated.}

\item{substitute}{Should \code{expr} be \code{\link[base:substitute]{base::substitute()}}:d or not.}

\item{threshold}{The smallest memory allocation (in bytes) to log.}

\item{on_error}{(character) Controls whether evaluation errors should
signal an error (\code{"error"}), a warning (\verb{"warning"), or be ignored (}"ignore"`; default).}
}
\value{
\code{profmem()} and \code{profmem_end()} returns the collected allocation
data as an \code{Rprofmem} data.frame with additional attributes set.
An \code{Rprofmem} data.frame has columns \code{what}, \code{bytes}, and \code{trace}, with:
\itemize{
\item \code{what}:  (character) type of memory event;
either \code{"alloc"} or \code{"new page"}
\item \code{bytes}: (numeric) number of bytes allocated or \code{NA_real_}
(when \code{what} is \code{"new page"})
\item \code{trace}: (list of character vectors) zero or more function names
}

The attributes set are:
\itemize{
\item \code{threshold} : The threshold used (= argument \code{threshold})
\item \code{expression}: The expression profiled (= argument \code{expr})
\item \code{value}     : The value of the evaluated expression
(only set if there was no error)
\item \code{error}     : The error object in case the evaluation failed
(only set if there was an error)
}

\code{profmem_begin()} returns (invisibly) the number of nested profmem
session currently active.

\code{profmem_suspend()} and \code{profmem_resume()} returns nothing.

\code{profmem_status()} returns \code{"inactive"}, \code{"active"},
or \code{"suspended"}.

\code{promem_depth()} returns a non-negative integer.
}
\description{
\code{profmem()} evaluates and memory profiles an \R expression.

\code{profmem_begin()} starts the memory profiling of all the following \R
evaluations until \code{profmem_end()} is called.
}
\details{
In order for memory profiling to work, \R must have been \emph{built} with memory
profiling enabled.  Function
\code{\link[base:capabilities]{base::capabilities("profmem")}} will
return \code{TRUE} of it is enabled, otherwise \code{FALSE}.
If memory profiling is \emph{not} supported, \code{profmem()} and \code{profmem_begin()}
will produce an informative error.  The pre-built \R binaries on
\href{https://cran.r-project.org/}{CRAN} support memory profiling.

What is logged?  The \code{profmem()} function uses \code{\link[utils:Rprofmem]{utils::Rprofmem()}} for
logging memory, which logs all memory \emph{allocations} that are done via the
\R framework.  Specifically, the logger is tied to \code{allocVector3()} part
of \R's native API.  This means that nearly all memory allocations done
in \R are logged. \emph{Neither} memory deallocations \emph{nor} garbage collection
events are logged.  Furthermore, allocations done by non-\R native libraries
or \R packages that use native code \code{Calloc() / Free()} for internal objects
are also \emph{not} logged.

Any memory events that would occur due to calling any of the \pkg{profmem}
functions themselves will \emph{not} be logged and \emph{not} be part of the returned
profile data (regardless whether memory profiling is active or not).
This is intentional.

If a profmem profiling is already active, \code{profmem()} and \code{profmem_begin()}
performs an \emph{independent}, \emph{nested} profiling, which does not affect the
already active one.  When the active one completes, it will contain all
memory events also collected by the nested profiling as if the nested one
never occurred.

Profiling gathered by \pkg{profmem} \emph{will} be corrupted if the code profiled
calls \code{\link[utils:Rprofmem]{utils::Rprofmem()}}, with the exception of such calls done via the
\pkg{profmem} package itself.
}
\examples{
if (capabilities("profmem")) {

## Memory profile an R expression
p <- profmem({
  x <- raw(1000)
  A <- matrix(rnorm(100), ncol = 10)
})

## Display the results
print(p)

## Total amount of memory allocation
total(p)

## Allocations greater than 1 kB
p2 <- subset(p, bytes > 1000)
print(p2)

## The expression is evaluated in the calling environment
str(x)
str(A)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readRprofmem.R
\name{readRprofmem}
\alias{readRprofmem}
\title{Read an Rprofmem log file}
\usage{
readRprofmem(pathname, as = c("Rprofmem", "fixed", "raw"), drop = 0L, ...)
}
\arguments{
\item{pathname}{The Rprofmem log file to be read.}

\item{as}{Specifies in what format data should be returned.
If \code{"raw"}, the line content of the file is returned as is
(as a character vector).
If \code{"fixed"}, as \code{"raw"} but with missing newlines added to lines with empty
stack calls that may be introduced in R (< 3.5.0) (see Ref. 1).
If \code{"Rprofmem"}, the collected Rprofmem data is fully
parsed into bytes and call stack information.}

\item{drop}{Number of levels to drop from the top of the call stack.}

\item{...}{Not used}
}
\value{
An \code{Rprofmem} data.frame or a character vector (if \code{as} is \code{"raw"}
or \code{"fixed"}).
An \code{Rprofmem} data.frame has columns \code{what}, \code{bytes}, and \code{trace}, with:
\itemize{
\item \code{what}:  (character) type of memory event;
either \code{"alloc"} or \code{"new page"}
\item \code{bytes}: (numeric) number of bytes allocated or \code{NA_real_}
(when \code{what} is \code{"new page"})
\item \code{trace}: (list of character vectors) zero or more function names
}
}
\description{
Reads and parses an Rprofmem log file that was created by
\code{\link[utils:Rprofmem]{utils::Rprofmem()}}.
}
\examples{
file <- system.file("extdata", "example.Rprofmem.out", package = "profmem")

raw <- readRprofmem(file, as = "raw")
cat(raw, sep = "\n")

profmem <- readRprofmem(file, as = "Rprofmem")
print(profmem)
}
\references{
Ref. 1: \url{https://github.com/HenrikBengtsson/Wishlist-for-R/issues/25}
}
